package listFW.factory;

import listFW.*;

/**
 * Manufactures concrete IMTList and INEList objects.  Has only one
 * instance referenced by CompositeListFactory.Singleton.
 * MTList and NEList are anonymous inner classes and hidden from all external
 * client code.  The implementations for MTList and NEList are the same as
 * before but completely invisible to the outside of this factory.
 * @author Dung X. Nguyen
 * @since Copyright 2004 by DXN - All rights reserved
 */
public class InnerCompListFactory implements IListFactory {
  
  /**
   * Note the use of private static final.
   */
  private static final IMTList MTList = new IMTList() {
    
    final public Object execute(IListAlgo algo, Object... inp) {
      return algo.emptyCase(this, inp);
    }
    
    public String toString() {
      return "()";
    }
  };
  
  /**
   * Singleton Pattern
   */
  public static final InnerCompListFactory Singleton = new InnerCompListFactory();
  private InnerCompListFactory() {
  }
  
  /**
   * Creates an empty list.
   * @return an IMTList object.
   */
  public IMTList makeEmptyList() {
    return MTList;
  }
  
  /**
   * Creates a non-empty list containing a given first and a given rest.
   * @param first a data object.
   * @param rest != null, the rest of the non-empty list to be manufactured.
   * @return an INEList object containing first and rest
   */
  public INEList makeNEList(final Object first, final IList rest) {
    return new INEList(){
      
      final public Object getFirst() {
        return first;
      }
      
      final public IList getRest() {
        return rest;
      }
      
      final public Object execute(IListAlgo algo, Object... inp) {
        return algo.nonEmptyCase(this, inp);
      }
      
      
      public String toString() {
        return (String)ToStringAlgo.Singleton.nonEmptyCase(this);
      }
    };
  }
}