package ring;
/**
 * Represents a doubly-connected mutable ring buffer
 */
public class RingBuffer {
  /**
   * Abstract state of the ring
   */
  private static abstract class ANode {
    /**
     * Get the next node in the ring
     */
    abstract ANode getNext();
    
    /**
     * Get the previous node in the ring
     */
    abstract ANode getPrev();
    
    /**
     * Get the data in this node
     */
    abstract Object getData();
    
    /**
     * Set the data in this node
     */
    abstract RingBuffer setData(Object data);
    
    /**
     * Insert data, pushing the current data 
     * towards the next node.
     */
    abstract RingBuffer insertFront(Object data);
    
    /**
     * Remove the data on the node, moving the data 
     * in from the next node.
     */
    abstract Object removeFront();
  }
  
  /**
   * Represents an abstract non-empty node
   */
  private abstract class ANENode extends ANode {
    /**
     * The data held at this node
     */
    protected Object _data;  
    
    /**
     * The next node
     */
    protected ANode _next= this;
    
    /**
     * The previous node
     */
    protected ANode _prev = this;
    
    /**
     * Constructor to create a node for a ring 
     * with single non-empty node
     */
    ANENode(Object data) {
      _data = data;
    }
    
    /**
     * Constructor to create a node with a given 
     * next and previous nodes.
     */
    ANENode(Object data, ANode next, ANode prev) {
      _next = next;
      _data = data;
      _prev = prev;
    }
    
    /**
     * Get the next node
     */
    ANode getNext() {
      return _next;
    }
    
    /**
     * Set the next node
     */
    void setNext(ANode next){
      _next = next;
    }
   
    
    /**
     * Get the previous node
     */
    ANode getPrev() {
      return _prev;
    }
    
    /**
     * Set the previous node
     */
    void setPrev(ANode prev) {
      _prev = prev;
    }
    /**
     * Get the data stored at this node
     */
    Object getData() {
      return _data;
    }
    
    /**
     * Sets the current data stored at this node
     * Returns the ringbuffer object.
     */
    RingBuffer setData(Object data) {
      _data = data;
      return RingBuffer.this;
    }
    
    /**
     * Abstract insertion process.  Depends on whether 
     * there are one or more elements in the ring.
     */
    abstract RingBuffer insertFront(Object data); 

    /**
     * Abstract removal process.  Depends on whether 
     * there are one or more elements in the ring.
     */
    abstract Object removeFront();
    
    /**
     * String representation of the ring.
     */
    public String toString() {
      String result = _data.toString()+"->";
      ANENode node = (ANENode) _next;
      while(node != this) {
        result += node.getData().toString()+"->";
        node = (ANENode) node.getNext();
      }
      return result+"["+_data.toString()+"]";
    }
  }
  
  /**
   * Represents the node of a single-node ring.
   */
  private class SingleNENode extends ANENode {
    /**
     * Instantiates one node for a single element ring
     */
    SingleNENode(Object data) {
      super(data);
    }
    
    /**
     * Insert data, pushing the current data 
     * towards the next node.
     * Transitions to a multi-node ring.
     */
    RingBuffer insertFront(Object data) {
      _state = new MultiNENode(data, _data); 
      return RingBuffer.this;
    }
    
    /**
     * Transition to the empty ring.
     * @return The original data held at this node
     */
    Object removeFront() {    
      _state = _emptyNode;
      return _data;
    }
  }
  
  /**
   * Represents a node on the ring when there is more
   * than one node present.
   */
  private class MultiNENode extends ANENode {
    
    /**
     * Instantiates two nodes in a two element ring.
     * @param data1 The data element of the current node.
     * @param data2 The data element of the other node.
     */
    MultiNENode(Object data1, Object data2) {
      super(data1);
      MultiNENode other = new MultiNENode(data2, this, this);
      _next = other;
      _prev = other;
    }
    
    /**
     * Normal constructor that creates a single node and 
     * sets both next and previous nodes
     */
    MultiNENode(Object data, ANode next, ANode prev) {
      super(data, next, prev);
    }
    
    /**
     * Insert data, pushing the current data 
     * towards the next node.
     */
    RingBuffer insertFront(Object data) {

      MultiNENode temp = new MultiNENode(_data, _next, this);
      ((ANENode)_next).setPrev(temp);
      _next = temp;
      _data = data;  
      return RingBuffer.this;
    }
    
    /**
     * Remove the data on the node, moving the data 
     * in from the next node.
     * If this is a one-element ring, transition to the 
     * empty ring.
     */
    Object removeFront() {
      Object result = _data;
      // Check if in two-element ring
      if(_next == _prev) {
        _state = new SingleNENode(_next.getData());
      }
      else {
        _data = _next.getData(); 
        _next = _next.getNext(); 
        ((ANENode)_next)._prev = this;
      }
      return result;
    }
  }
  
  
  /**
   * Singleton empty node instance
   * Throws IllegalStateExceptions for all methods
   * except insertFront(), where it changes the RingBuffer's 
   * state to a single non-empty node and
   * getNext() and getPrev() which do nothing.
   */
  private final ANode _emptyNode = new ANode(){
    ANode getNext() {
      return this;
    }
    ANode getPrev() {
      return this;
    }
    Object getData() {
      throw new IllegalStateException("Empty buffer has no data.");
    }
    RingBuffer setData(Object data){
      throw new IllegalStateException("Empty buffer has no data.");
    }
    
    RingBuffer insertFront(Object data) {
      _state = new SingleNENode(data);
      return RingBuffer.this;
    }
    Object removeFront(){
      throw new IllegalStateException("Empty buffer has no data.");
    }
    
    public String toString(){
      return "";
    }
  };
  
  /**
   * The current state of the ring.  Initialized to an empty ring.
   */
  private ANode _state = _emptyNode;
  
  /**
   * The current number of elements in the ring.
   */
  private int _count = 0;
  
  /**
   * Insert a data element into the ring, in pushing the 
   * current data towards next.
   * The ring is returned.
   */
  public RingBuffer insertFront(Object data) {
    _count++;
    return _state.insertFront(data);
  }
  
  /**
   * Removes the current data element from the ring.
   * Data is moved in from the direction of next.
   * Returns the removed element value.
   */
  public Object removeFront() {
    _count--;
    return _state.removeFront();
  }
  
  /**
   * Get the current count of the number of elements 
   * in the ring.
   */
  public int getCount() {
    return _count;
  }
  
  /**
   * Get the data element from the current node 
   * in the ring.
   */
  public Object getFirst() {
    return _state.getData();
  }
  
  /**
   * Set the given data element into the current node 
   * in the ring.
   */
  public RingBuffer setFirst(Object data) {
    return _state.setData(data);
  }
  
  /**
   * Rotates the ring towards the next element.
   * The current node changes to the next node.
   */
  public RingBuffer toNext() {
    _state = _state.getNext();
    return this;
  }
  
  /**
   * Rotates the ring towards the previous element.
   * The current node changes to the previous node.
   */
  public RingBuffer toPrev() {
    _state = _state.getPrev();
    return this;
  }
  
  /**
   * Returns a String representation of the ring.  
   * The data elements are listed in order starting from 
   * the current node and following in the direction of 
   * next until the current node is encountered again.
   * The reoccurance of the current node is indicated by 
   * enclosing the data value in square brackets.
   * Sample output:  "{a->b->c->d->[a]}"
   */
  public String toString() {
    return "{"+_state.toString()+"}"; 
  }
}

